"""
### Handle nodata values in rasters. ###

A module to handle the various aspects of NODATA in raster files.

TODO:
    * raster_to_mask
    * raster_invert_nodata
"""

# Standard library
import sys; sys.path.append("../../")

# External
import numpy as np
from osgeo import gdal

# Internal
from buteo.utils import core_utils, gdal_utils, gdal_enums
from buteo.raster import core_raster



def raster_has_nodata_value(raster):
    """
    Check if a raster or a list of rasters contain nodata values

    ## Args:
    `raster` (_str_/_gdal.Dataset_, _list_): The raster to check for nodata values.

    ## Returns:
    (_bool_): **True** if the raster or list of rasters contain nodata values.
    """
    core_utils.type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")

    rasters = gdal_utils.get_path_from_dataset(core_utils.ensure_list(raster))
    assert gdal_utils.is_raster_list(raster), f"List contains invalid rasters. {raster}"

    nodata_values = []
    for internal_raster in rasters:
        raster_metadata = core_raster._raster_to_metadata(internal_raster)

        raster_nodata = raster_metadata["nodata_value"]

        if raster_nodata is not None:
            nodata_values.append(True)
        else:
            nodata_values.append(False)

    if isinstance(raster, list):
        return nodata_values

    return nodata_values[0]


def raster_get_nodata_value(raster):
    """
    Get the nodata value of a raster or a list of rasters.

    ## Args:
    `raster` (_str_/_gdal.Dataset_, _list_): The raster(s) to get nodata values from.

    ## Returns:
    (_float_/_int_/_list_): The nodata value(s) of the raster(s).
    """
    core_utils.type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")

    rasters = gdal_utils.get_path_from_dataset(core_utils.ensure_list(raster))
    assert gdal_utils.is_raster_list(raster), f"List contains invalid rasters. {raster}"

    nodata_values = []
    for internal_raster in rasters:
        raster_metadata = core_raster._raster_to_metadata(internal_raster)

        raster_nodata = raster_metadata["nodata_value"]

        nodata_values.append(raster_nodata)

    if isinstance(raster, list):
        return nodata_values
    else:
        return nodata_values[0]


def raster_set_nodata(
    raster,
    dst_nodata,
    out_path=None,
    *,
    overwrite=True,
    in_place=False,
    prefix="",
    suffix="_nodata_set",
    creation_options=None,
):
    """
    Sets all the nodata for raster(s) to a value.

    ## Args:
    `raster` (_str_/_gdal.Dataset_/_list_): The raster(s) to set nodata values for. </br>
    `dst_nodata` (_float_/_int_/_str_/_None_): The target nodata value. If 'infer' the nodata
        value is set based on the input datatype. A list of nodata values can be based matching
        the amount of input rasters. If multiple nodata values should be set,
        use raster_mask_values. </br>

    ## Kwargs:
    `out_path` (_str_/_list_/_None_): The destination of the changed rasters. (Default: **None**) </br>
    `overwrite` (_bool_): Should the rasters be overwritten if they already exist? (Default: **True**) </br>
    `in_place` (_bool_): Should the rasters be changed in_place or copied? (Default: **False**) </br>
    `prefix` (_str_): Prefix to add to the output. (Default: **""**) </br>
    `suffix` (_str): Suffix to add to the output. (Default: **""**) </br>
    `creation_options` (_list_): Creation options for the output rasters. (Default: **None**) </br>

    ## Returns:
    (_str_/_list_): Returns the rasters with nodata set. If in_place is True a reference to the
    changed orignal is returned, otherwise a copied memory raster or the path to the
    generated raster is outputted.
    """
    core_utils.type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    core_utils.type_check(dst_nodata, [float, int, str, list, None], "dst_nodata")
    core_utils.type_check(out_path, [list, str, None], "out_path")
    core_utils.type_check(overwrite, [bool], "overwrite")
    core_utils.type_check(prefix, [str], "prefix")
    core_utils.type_check(suffix, [str], "postfix")
    core_utils.type_check(creation_options, [[str], None], "creation_options")

    raster_list = core_utils.ensure_list(raster)
    path_list = gdal_utils.create_output_path_list(
        raster_list,
        out_path,
        prefix=prefix,
        suffix=suffix,
        overwrite=overwrite,
    )

    rasters_metadata = []
    internal_dst_nodata = None

    if isinstance(dst_nodata, str) and dst_nodata != "infer":
        raise ValueError(f"Invalid dst_nodata value. {dst_nodata}")

    if isinstance(dst_nodata, list):
        if not isinstance(raster, list) or len(dst_nodata) != len(raster):
            raise ValueError(
                "If dst_nodata is a list, raster must also be a list of equal length."
            )

        for value in dst_nodata:
            if isinstance(value, (float, int, str, None)):
                raise ValueError("Invalid type in dst_nodata list.")

            if isinstance(value, str) and value != "infer":
                raise ValueError("If dst_nodata is a string it must be 'infer'")

    output_rasters = []

    for index, internal_raster in enumerate(raster_list):

        raster_metadata = None
        if len(rasters_metadata) == 0:
            raster_metadata = core_raster._raster_to_metadata(internal_raster)

            if not isinstance(raster_metadata, dict):
                raise Exception("Metadata is in the wrong format.")

            rasters_metadata.append(raster_metadata)
        else:
            raster_metadata = rasters_metadata[index]

        if dst_nodata == "infer":
            internal_dst_nodata = gdal_enums.translate_gdal_dtype_to_str(
                raster_metadata["dtype_gdal_raw"]
            )
        elif isinstance(dst_nodata, list):
            internal_dst_nodata = dst_nodata[index]
        else:
            internal_dst_nodata = dst_nodata

        if in_place:
            for band in range(raster_metadata["bands"]):
                raster_band = internal_raster.GetRasterBand(band + 1)
                raster_band.SetNodataValue(internal_dst_nodata)
                raster_band = None
        else:
            raster_mem = None

            if out_path is None:
                raster_mem = gdal_utils.save_dataset_to_memory(internal_raster)
            else:
                core_utils.remove_if_required(path_list[index], overwrite)
                raster_mem = gdal_utils.save_dataset_to_disk(internal_raster, path_list[index], creation_options=creation_options)

            for band in range(raster_metadata["bands"]):
                raster_mem_ref = core_raster._open_raster(raster_mem)
                raster_band = raster_mem_ref.GetRasterBand(band + 1)
                raster_band.SetNodataValue(internal_dst_nodata)

    if isinstance(raster, list):
        return output_rasters

    return output_rasters[0]


def raster_remove_nodata(
    raster,
    out_path=None,
    *,
    overwrite=True,
    in_place=False,
    prefix="",
    suffix="_nodata_removed",
    creation_options=None,
):
    """
    Removes all the nodata from a raster or a list of rasters.

    ## Args:
    `raster` (_str_/_gdal.Dataset_/_list_): The raster(s) to remove nodata values for. </br>

    ## Kwargs:
    `out_path` (_str_/_list_/_None_): The destination of the changed rasters. (Default: **None**) </br>
    `overwrite` (_bool_): Should the rasters be overwritten if they already exist? (Default: **True**) </br>
    `in_place` (_bool_): Should the rasters be changed in_place or copied? (Default: **False**) </br>
    `prefix` (_str_): Prefix to add to the output. (Default: **""**) </br>
    `suffix` (_str_): Suffix to add to the output. (Default: **"_nodata_removed"**) </br>
    `creation_options` (_list_): Creation options for the output rasters. (Default: **None**) </br>

    ## Returns:
    (_str_/_list_): Returns the rasters with nodata removed. If in_place is True a reference to the
    changed orignal is returned, otherwise a copied memory raster or the path to the
    generated raster is outputted.
    """

    return raster_set_nodata(
        raster,
        dst_nodata=None,
        out_path=out_path,
        in_place=in_place,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        creation_options=creation_options,
    )


def raster_mask_values(
    raster,
    values_to_mask,
    out_path=None,
    *,
    include_original_nodata=True,
    dst_nodata="infer",
    in_place=False,
    overwrite=True,
    prefix="",
    suffix="",
    add_uuid=False,
    creation_options=[],
):
    """
    Mask a raster with a list of values.

    ## Args:
    `raster` (_str_/_gdal.Dataset_/_list_): The raster(s) to mask. </br>
    `values_to_mask` (_list_): The values to mask. </br>

    ## Kwargs:
    `include_original_nodata` (_bool_): Should the nodata_value of the input raster be added
    to the list of masked values? (Default: **True**) </br>
    `dst_nodata` (_float_/_int_/_str_/_list_): The nodata value to use for the output raster.
    If infer, the nodata_value from the input raster is used. (Default: **"infer"**) </br>
    `out_path` (_str_/_list_/_None_): The destination of the changed rasters.
    If out_paths are specified, in_place is automatically set to False. The path can be a folder. (Default: **None**)</br>
    `in_place` (_bool_): Should the rasters be changed in_place or copied? (Default: **False**) </br>
    `overwrite` (_bool_): If the output path exists already, should it be overwritten? (Default: **True**)</br>
    `prefix` (_str_): Prefix to add to the output. (Default: **""**) </br>
    `suffix` (_str_): Suffix to add to the output. (Default: **""**) </br>
    `add_uuid` (_bool_): Should a uuid be added to the output path? (Default: **False**) </br>
    `creation_options` (_list_/_None_): The GDAL creation options to be passed. (Default: **None**) </br>

    Returns:
    (_str_/_list_): Returns the rasters with nodata removed. If in_place is True a reference to the
    changed orignal is returned, otherwise a copied memory raster or the path to the
    generated raster is outputted.
    """
    core_utils.type_check(raster, [str, gdal.Dataset, [str, gdal.Dataset]], "raster")
    core_utils.type_check(values_to_mask, [[int, float, None]], "values_to_mask")
    core_utils.type_check(out_path, [[str], str, None], "out_path")
    core_utils.type_check(include_original_nodata, [bool], "include_original_nodata")
    core_utils.type_check(dst_nodata, [float, int, str, [float, int, str, None], None], "dst_nodata")
    core_utils.type_check(in_place, [bool], "in_place")
    core_utils.type_check(overwrite, [bool], "overwrite")
    core_utils.type_check(prefix, [str], "prefix")
    core_utils.type_check(suffix, [str], "postfix")
    core_utils.type_check(creation_options, [[str], None], "creation_options")

    rasters_metadata = []
    internal_in_place = in_place if out_path is None else False
    internal_dst_nodata = None

    for value in values_to_mask:
        if not isinstance(value, (int, float)):
            raise ValueError("Values in values_to_mask must be ints or floats")

    if isinstance(dst_nodata, str) and dst_nodata != "infer":
        raise ValueError(f"Invalid dst_nodata value. {dst_nodata}")

    if isinstance(dst_nodata, list):
        if not isinstance(raster, list) or len(dst_nodata) != len(raster):
            raise ValueError(
                "If dst_nodata is a list, raster must also be a list of equal length."
            )

        for value in dst_nodata:
            if isinstance(value, (float, int, str, None)):
                raise ValueError("Invalid type in dst_nodata list.")

            if isinstance(value, str) and value != "infer":
                raise ValueError("If dst_nodata is a string it must be 'infer'")

    raster_list = core_utils.ensure_list(raster)
    out_paths = gdal_utils.create_output_path_list(
        raster_list,
        out_path=out_path,
        overwrite=overwrite,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
    )


    output_rasters = []

    for index, internal_raster in enumerate(raster_list):

        raster_metadata = None
        if len(rasters_metadata) == 0:
            raster_metadata = core_raster._raster_to_metadata(internal_raster)
            rasters_metadata.append(raster_metadata)
        else:
            raster_metadata = rasters_metadata[index]

        if dst_nodata == "infer":
            internal_dst_nodata = gdal_enums.translate_gdal_dtype_to_str(
                raster_metadata["dtype_gdal_raw"]
            )
        elif isinstance(dst_nodata, list):
            internal_dst_nodata = dst_nodata[index]
        else:
            internal_dst_nodata = dst_nodata

        mask_values = list(values_to_mask)
        if include_original_nodata:
            if raster_metadata["nodata_value"] is not None:
                mask_values.append(raster_metadata["nodata_value"])

        arr = core_raster.raster_to_array(internal_raster, filled=True)

        mask = None
        for idx, mask_value in enumerate(mask_values):
            if idx == 0:
                mask = arr == mask_value
            else:
                mask = mask/arr == mask_value

        arr = np.ma.masked_array(arr, mask=mask, fill_value=internal_dst_nodata)

        if internal_in_place:
            for band in range(raster_metadata["bands"]):
                raster_band = internal_raster.GetRasterBand(band + 1)
                raster_band.WriteArray(arr[:, :, band])
                raster_band = None
        else:
            out_name = out_paths[index]
            core_utils.remove_if_required(out_name, overwrite)

            output_rasters.append(
                core_raster.array_to_raster(arr, internal_raster, out_path=out_name)
            )

    if isinstance(raster, list):
        return output_rasters

    return output_rasters[0]



"""
### Reproject vectors. ###

Functions to reproject vectors. References can be both vector and raster.
"""

# Standard library
import sys; sys.path.append("../../")
import os

# External
import osgeo
from osgeo import gdal, osr, ogr

# Internal
from buteo.utils import gdal_utils, core_utils
from buteo.vector import core_vector



def _reproject_vector(
    vector,
    projection,
    out_path=None,
    *,
    copy_if_same=False,
    overwrite=True,
    prefix="",
    suffix="",
):
    """ Internal. """
    assert isinstance(vector, (ogr.DataSource, str)), "Invalid vector input"
    assert gdal_utils.is_vector(vector), "Invalid vector input"

    origin = core_vector._open_vector(vector)
    out_name = out_path
    if out_path is None:
        out_name = gdal_utils.create_memory_path(
            gdal_utils.get_path_from_dataset(vector),
            prefix=prefix,
            suffix=suffix,
            add_uuid=True,
        )

    metadata = core_vector._vector_to_metadata(origin)
    namesplit = os.path.splitext(os.path.basename(out_name))
    out_name = os.path.join(os.path.dirname(out_name), prefix + namesplit[0] + suffix + namesplit[1])

    origin_projection = metadata["projection_osr"]
    target_projection = gdal_utils.parse_projection(projection)

    if not isinstance(target_projection, osr.SpatialReference):
        raise Exception("Error ")

    if origin_projection.IsSame(target_projection):
        if copy_if_same:
            if out_path is None:
                return gdal_utils.save_dataset_to_memory(origin)

            return gdal_utils.save_dataset_to_disk(origin, out_name)
        else:
            return gdal_utils.get_path_from_dataset(vector)

    # GDAL 3 changes axis order: https://github.com/OSGeo/gdal/issues/1546
    if int(osgeo.__version__[0]) >= 3:
        origin_projection.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)
        target_projection.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

    coord_trans = osr.CoordinateTransformation(origin_projection, target_projection)

    core_utils.remove_if_required(out_path, overwrite)

    driver = ogr.GetDriverByName(gdal_utils.path_to_driver_vector(out_name))
    destination: ogr.DataSource = driver.CreateDataSource(out_name)

    for layer_idx in range(len(metadata["layers"])):
        origin_layer = origin.GetLayerByIndex(layer_idx)
        origin_layer_defn = origin_layer.GetLayerDefn()

        layer_dict = metadata["layers"][layer_idx]
        layer_name = layer_dict["layer_name"]
        layer_geom_type = layer_dict["geom_type_ogr"]

        destination_layer = destination.CreateLayer(
            layer_name, target_projection, layer_geom_type
        )
        destination_layer_defn = destination_layer.GetLayerDefn()

        # Copy field definitions
        origin_layer_defn = origin_layer.GetLayerDefn()
        for i in range(0, origin_layer_defn.GetFieldCount()):
            field_defn = origin_layer_defn.GetFieldDefn(i)
            destination_layer.CreateField(field_defn)

        # Loop through the input features
        for _ in range(origin_layer.GetFeatureCount()):
            feature = origin_layer.GetNextFeature()
            geom = feature.GetGeometryRef()
            geom.Transform(coord_trans)

            new_feature = ogr.Feature(destination_layer_defn)
            new_feature.SetGeometry(geom)

            # Copy field values
            for i in range(0, destination_layer_defn.GetFieldCount()):
                new_feature.SetField(
                    destination_layer_defn.GetFieldDefn(i).GetNameRef(),
                    feature.GetField(i),
                )

            destination_layer.CreateFeature(new_feature)

        destination_layer.ResetReading()
        destination_layer = None

    destination.FlushCache()

    return out_name


def reproject_vector(
    vector,
    projection,
    out_path=None,
    *,
    copy_if_same=False,
    prefix="",
    suffix="",
    add_uuid=False,
    allow_lists=True,
    overwrite=True,
):
    """Reprojects a vector given a target projection.

    Args:
        vector (path/vector): The vector to reproject.

        projection (str/int/vector/raster): The projection is infered from
        the input. The input can be: WKT proj, EPSG proj, Proj, or read from a
        vector or raster datasource either from path or in-memory.

    **kwargs:
        out_path (path/None): The destination to save to. If None then
        the output is an in-memory raster.

        overwite (bool): Is it possible to overwrite the out_path if it exists.

    Returns:
        An in-memory vector. If an out_path is given, the output is a string containing
        the path to the newly created vecotr.
    """
    core_utils.type_check(vector, [str, ogr.DataSource], "vector")
    core_utils.type_check(projection, [str, int, ogr.DataSource, gdal.Dataset, osr.SpatialReference], "projection")
    core_utils.type_check(out_path, [str, [str], None], "out_path")
    core_utils.type_check(copy_if_same, [bool], "copy_if_same")
    core_utils.type_check(prefix, [str], "prefix")
    core_utils.type_check(suffix, [str], "suffix")
    core_utils.type_check(add_uuid, [bool], "add_uuid")
    core_utils.type_check(allow_lists, [bool], "allow_lists")
    core_utils.type_check(overwrite, [bool], "overwrite")

    if not allow_lists and isinstance(vector, list):
        raise ValueError("Lists are not allowed when allow_lists is False.")

    vector_list = core_utils.ensure_list(vector)

    assert gdal_utils.is_vector_list(vector_list), f"Invalid input vector: {vector_list}"

    path_list = gdal_utils.create_output_path_list(
        vector_list,
        out_path=out_path,
        prefix=prefix,
        suffix=suffix,
        add_uuid=add_uuid,
    )

    assert core_utils.is_valid_output_paths(path_list, overwrite=overwrite), "Invalid output path generated."

    output = []
    for index, in_vector in enumerate(vector_list):
        output.append(
            _reproject_vector(
                in_vector,
                projection,
                out_path=path_list[index],
                copy_if_same=copy_if_same,
                overwrite=overwrite,
                prefix=prefix,
                suffix=suffix,
            )
        )

    if isinstance(vector, list):
        return output

    return output[0]
